# SecMonitor

Proj118: An eBPF-based Monitor for Container 基于eBPF技术的容器行为监控工具

New York University Shanghai 上海纽约大学

NYUSHTeam1 上海纽约大学1队 门秋彤

指导教师: 刘古月 程泽睿智

---

## [开发文档](/assets/devdoc.md)

## Introduction

SecMonitor is a concise eBPF-based Monitor for Container, which **eliminates the dependence** on specific kernel version and syscall type (though must supports eBPF). As a dynamic analysis tool, SecMonitor cannot ensure 100% accuracy of correct minimum previleges set. So we also decoupled the Tester as an independent thread, which brings potential flexibility of creating customized test scripts based on container types and functions.

### Key Features

- Suitable for multi-threaded containers, based on the cgroup property.
- Concise BPF C code reducing kernel overheads, using an abstract kernel tracepoint: SYS_ENTER.
- Flexible test scripts increasing the accuracy of previleges set.

## Architecture

![arch](/assets/archu.svg)

The architecture of SecMonitor. 

- Docker Client utilizes Docker Python SDK to launch certain container and reports information of processes inside the container. 
- Tester runs as an independent thread and execute proper test scripts to stimulate the container, making it perform as many system calls as possible.
- Monitor loads the BPF C code into the eBPF environment in the kernel through BCC, and consumes the output of our BPF Program through BCC's API. It produces a statistics file after the monitoring is stopped.
- Analyzer reads the default SECCOMP configuration, a syscall table of current kernel version, and the statistics file produced by Monitor, and generates a secure and accurate SECCOMP Config File.

## Usage

### Dependency

python packages:

```
docker-py
bcc
```

system requirements:

```
Docker
linux kernel >= 5.0.0
```

### Run

```bash
./monitor.py -i|--image <image-name> [-d|--display] [-t|--test <test_script.py>] [-c|--config <docker_config>] [-l|--log <log_address>] [-o|--output <statistics_output_address>] [-g|--generate [config_generation_address]]
```

- ```-i```: Required. Setting docker image name to run and monitor.
- ```-d```: Optional. Whether display every captured syscall records into logs. Default is True.
- ```-t```: Optional. Setting customized test script to run.
- ```-c```: Optional. Setting docker configuration, a json file.
- ```-l```: Optional. Setting log file address, default secmon.log
- ```-o```: Optional. Setting statistics file output address.
- ```-g```: Optional. Whether generate the SECCOMP config file for the monitored container. One may use ```generate.py``` later to do this.

```bash
./generate.py <statistics_file> [-o|--output <config_generation_address>]
```

- ```statistics_file```: Required. Setting the statistics file to be analyzed.
- ```-o```: Optional. Setting SECCOMP config file output address. Default address is the same as statistics file, named ```<statistics_file>_secmonitor.json```

```bash
./arch.py
```

- Generating syscalls id-name mapping table by kernel version and architecture.

## Examples

The following example monitors the behavior of a plain nginx container when responding to a GET request.

### Operations

- test\_case\_1.py:

```python
import requests

ret = requests.get("http://127.0.0.1:8080/")
print(ret.text)
```

![ops](/assets/terminal.png)

### Results

- syscalls.json

![syscalls](/assets/statistics.png)

- secmon.log

![log](/assets/log.png)

- secmonitor.json

![log](/assets/seccomp.png)

Compared with the default whitelist in [the default SECCOMP config](https://github.com/moby/moby/blob/master/profiles/seccomp/default.json), we have decreased the number from 200+ to only 21.

The following example shows how to generate a syscall id-name mapping table using `arch.py`. (On Linux 5.14, x86 platform.)

### Operations

![archg](/assets/arch.png)

### Results

![tbl](/assets/table.png)
